//=============================================================================
//
class MAS14_MK_PID extends MAS14_Actor ClassGroup(MAS14) placeable;
//
// CBP 17-02-21 Explores PID controller applied to mass tethered with springs
// Generates Octave .m file to plot data and compare with theory (using the
// 'transfer function' approach
//
// Status: RELEASE
//
//=============================================================================

//static mesh components that make up this model--------------------------------

var StaticMeshComponent airtrack;
var StaticMeshComponent glider;
var StaticMeshComponent springLeft;
var StaticMeshComponent springRight;
var StaticMeshComponent clampLeft;
var StaticMeshComponent clampRight;

// Editable variables ----------------------------------------------------------
// These are normally the parameters of the model and the model's initial
// conditions. These should have default properties assigned.

var() float mmass;
var() float initX;
var() float initVx;
var() float stiffnessX;
var() float damping;
var() float desX;

// Integrator control
var() float deltaT;
var int itn;

// PID Parameters
var() float Kp;
var() float Ki;
var() float Kd;

var float e1,e2,u,du;


// Internal variables ----------------------------------------------------------
// Internal "state" variables.
var float time;
var float velyX;
var float dispX;
var float error;
var float forceX;

// Internal Constants. These may be used for conversions between units.
var float scaling;

var FileLog mlog;         // Matlab file.

// functions********************************************************************

// Initialisation of this actor. do not change this. ---------------------------

function Actor_Initialize(PlayerController newPlayer){

  bRunning = false;

  super.Actor_Initialize(newPlayer);

  openMatlabFile();

  setLogFileColumnLabels();
  initializeVariables();
  setInitialConditions();
  visualization();

  setTimer(deltaT,true,'computeTimer');

}

// Sets the initial values for this actor. You will need to write this code ----

function setInitialConditions() {
  time = 0;
  dispX = initX;
}

// Resets the actor to its initial location and/or rotation and resets the
// initial values of the "state" variables.

function reset() {
  initializeVariables();         // set the variables to zeros
  setInitialConditions();        // set init conds of variables which need it
  visualization();

}

// Here you must set all internal "state" variables to zero.

function initializeVariables() {

  time = 0;
  forceX = 0.0;
  u = 0;
  error = 0;
  e1 = 0;
  e2 = 0;

}
// Sets the column labels in the data log file. One line of code for each label-

function setLogFileColumnLabels() {
  local array<String> columnLabels;
  columnLabels.length = 0; //empty array.
  columnLabels[0] = "time";
  columnLabels[1] = "dispX";
  writeMatlabFileHeader(time,dispX);
}

// Here you specify which VARIABLES will be logged. This should agree with your
// column labels.

function logDataRecord() {
  local array<float> dataArray;
  dataArray.length = 0; //empty array.
  dataArray[0] = time;
  dataArray[1] = dispX;
  writeMatlabFileRecord(time,dispX);
}

function logData() {
  local int m;
  m = int(logInterval/deltaT);
  if( itn % m == 0) logDataRecord();
}

// ============================ CVI ============================================

// called by the game engine on each rendering frame. Do not change this! Your
// changes to the code should appear in the computation() and visualization()
// functions.

function computeTimer(){ //-----------------------------------------------------

 if(bRunning) {
   computation(deltaT);
   visualization();
   time += deltaT;
   itn++;
   logData();   
 }

}

// COMPUTATION
// Here you will write your code based on your mathematical model --------------

function computation(float dT) {

  local float accelX;
  local float Ts;

  Ts = dT;

  // >>>>>>>>>>>>>>>> PUT YOUR PLANT CODE AFTER THIS LINE <<<<<<<<<<<<<<<<<<<<<<






  // >>>>>>>>>>>>>>>> PUT YOUR CONTROLLER CODE AFTER THIS LINE <<<<<<<<<<<<<<<<<






  
}

// VISUALISATION
// Here you will write your code to visualise your variables -------------------

function Visualization(){
  local vector newVector;

  newVector = default.glider.Translation;
  newVector.X = dispX * scaling;

  glider.SetTranslation(newVector);

  //visualize springs
  newVector = springLeft.Scale3D;
  newVector.X = (glider.Translation.X - clampleft.Translation.X)/216;
  springLeft.SetScale3D(newVector);
  newVector = springLeft.Translation;
  newVector.X = (glider.Translation.X + clampleft.Translation.X)/2;
  springLeft.SetTranslation(newVector);

  newVector = springRight.Translation;
  newVector.X = (glider.Translation.X + clampRight.Translation.X)/2;
  springRight.SetTranslation(newVector);
  newVector = springRight.Scale3D;
  newVector.X = (clampRight.Translation.X - glider.Translation.X)/216;
  springRight.SetScale3D(newVector);
}

// INTERACTION
// used to display PARAMETERS or VARIABLES on the HUD --------------------------

function SendValuesToHUDX( out array<string> HUDLine, out array<Color> LineCol){

  HUDLine.length = 0; //empty array.

  HUDLine[0] = "Running: "@bRunning;  // leave this alone
  HUDLine[1] = "Time: "@time;
  HUDLine[2] = "mmass: "@mmass;
  HUDLine[3] = "stiffnessX: "@stiffnessX;
  HUDLine[4] = "damping: "@damping;
  HUDLine[5] = "desX: "@desX;
  HUDLine[6] = "dispX: "@dispX;
  HUDLine[7] = "velyX: "@velyX;
  HUDLine[8] = "error: "@error;
  HUDLine[9] = "deltaT: "@deltaT;
  HUDLine[10] = "Kp: "@Kp;
  HUDLine[11] = "Ki: "@Ki;
  HUDLine[12] = "Kd: "@Kd ;

  // Assign colors to each line on the HUD. No entry here will assign a default
  // value of light blue.
  LineCol.length = 0;
  LineCol[0] = MakeColor(255,255,0,255);
  LineCol[1] = MakeColor(255,0,0,255);
  LineCol[2] = MakeColor(0,0,255,255);
  LineCol[3] = MakeColor(0,255,255,255);
  LineCol[4] = MakeColor(0,255,255,255);
  LineCol[5] = MakeColor(0,255,255,255);
  LineCol[6] = MakeColor(0,255,255,255);

}

// INTERACTION
// Here you will write code to be able to change the values of parameters from
// the command line.

function SetParam(string paramName,float paramValue){
  WorldInfo.Game.Broadcast(self,paramName@paramValue);

  paramName = Caps(paramName); //capitalize so case insensitive

  if(paramName == Caps("Kp")){
    Kp = paramValue;
  }
  else if(paramName == Caps("Ki")){
    Ki = paramValue;
  }
  else if(paramName == Caps("Kd")){
    Kd = paramValue;
  }
  else if(paramName == Caps("deltaT")){
    deltaT = paramValue;
    setTimer(deltaT,true);
  }
  else if(paramName == Caps("damping")){
    damping = paramValue;
  }
  else if(paramName == Caps("desX")){
    desX = paramValue;
  }
  else if(paramName == Caps("mmass")){
    mmass = paramValue;
  }
  else if(paramName == Caps("stiffnessX")){
    stiffnessX = paramValue;
  }
}

//process key presses for this actor--------------------------------------------
// Here you can assign various keys to produce any action you like. Each key is
// associated with a function, or global variable.

function ProcessKey(int controllerID,name key,EInputEvent eventType){
  WorldInfo.Game.Broadcast(self,"key pressed was:"@key);  //displays key pressed

  if(eventType == IE_Released){
    if(Key == 'F1'){    //toggles bRunning (global variable)
      bRunning = !bRunning;
    }
    if(Key == 'F2'){
	  bRunning = false;
      closeMatlabFile();
    }
    if(Key == 'F4') {
	  bRunning = false;
      closeMatlabFile();
    }
    if(Key == 'F12') {
	  bRunning = false;
      closeMatlabFile();
    }	
  }
}

// Helper Functions for this experiment Actor **********************************

 // ------------------- Matlab File Interface -----------------------------------

function openMatlabFile() {
  // set up the logging files.
  mlog = spawn(class'FileLog');
  mlog.Openlog(logFilename,".m",true);   // true means overwrite, else false.
}

function writeMatlabFileHeader(float p1, float p2) {
  mlog.Logf("time = "$p1);
  mlog.Logf("dispX = "$p2);
  mlog.Logf("");
  mlog.logF("A=[");
}

function writeMatlabFileRecord(float x, float y) {
  mlog.logF(x$","$y);
}

function writeMatlabFileFooter(float mmass, float damping, float stiffnessX, float Kp, float Ki, float Kd) {
  
  mlog.logF("];");
  mlog.logF("x = A(:,1);");
  mlog.logF("y = A(:,2);");
  mlog.logF("plot(x,y,'o');");
  mlog.logF("hold on");
  
  // Here comes the transfer function calculation
  // First open loop then closed loop
  mlog.logF("mmass = "$mmass);
  mlog.logF("damping = "$damping);
  mlog.logF("stiffnessX = "$stiffnessX);
  mlog.logF("Kp = "$Kp);
  mlog.logF("Ki = "$Ki);
  mlog.logF("Kd = "$Kd);  
	
  if(Kp == 0) {
    mlog.logF("num = [1];");
    mlog.logF("den = [mmass damping stiffnessX];");
    mlog.logF("sys =tf(num,den);");
    mlog.logF("step(sys,max(x));");    
  } else {
    mlog.logF("num = [Kd Kp Ki];");
    mlog.logF("den = [mmass damping+Kd stiffnessX+Kp Ki];");
    mlog.logF("sys =tf(num,den);");
    mlog.logF("step(sys,max(x));");  
  }
  mlog.logF("title('UDK data (circles) and theoretical response (line)');");
  mlog.logF("ylabel('displacement');");
  mlog.logF("hold off");

}

function closeMatlabFile() {
  writeMatlabFileFooter(mmass,damping,stiffnessX,Kp,Ki,Kd);
  mlog.Closelog();
}


defaultproperties  //***********************************************************
{

 bEdShouldSnap = true

  LightEnvironment=MyLightEnvironment  //need this to get lighting in UDK
  Components.Add(MyLightEnvironment)

  //base
   Begin Object class=StaticMeshComponent Name=StaticMeshComponent0
    StaticMesh=StaticMesh'NJS_LabAssets.MassSpringSystem.AirTrackShort'
	LightEnvironment=MyLightEnvironment
    bUsePrecomputedShadows=FALSE
  End Object
  Components.Add(StaticMeshComponent0)

  //glider
  Begin Object class=StaticMeshComponent Name=StaticMeshComponent1
    StaticMesh=StaticMesh'NJS_LabAssets.MassSpringSystem.AirTrackGlider'
	LightEnvironment=MyLightEnvironment
	bUsePrecomputedShadows=FALSE
  End Object
  Components.Add(StaticMeshComponent1)

  //left spring
  Begin Object class=StaticMeshComponent Name=StaticMeshComponent2
    StaticMesh=StaticMesh'NJS_LabAssets.MassSpringSystem.Spring'
	LightEnvironment=MyLightEnvironment
	bUsePrecomputedShadows=FALSE
	Translation=(X=-108,Y=0.0,Z=52.0)
	Scale3D=(X=1,Y=0.75,Z=0.75)
  End Object
  Components.Add(StaticMeshComponent2)

  //right spring
  Begin Object class=StaticMeshComponent Name=StaticMeshComponent3
    StaticMesh=StaticMesh'NJS_LabAssets.MassSpringSystem.Spring'
	LightEnvironment=MyLightEnvironment
	bUsePrecomputedShadows=FALSE
	Translation=(X=108,Y=0.0,Z=52)
	Scale3D=(X=1,Y=0.75,Z=0.75)
  End Object
  Components.Add(StaticMeshComponent3)

    //left clamp
  Begin Object class=StaticMeshComponent Name=StaticMeshComponent4
    StaticMesh=StaticMesh'NJS_LabAssets.MassSpringSystem.Clamp'
	LightEnvironment=MyLightEnvironment
	bUsePrecomputedShadows=FALSE
	scale = 0.5
    Translation=(X=-216.0,Y=0.0,Z=52.0)
    Rotation=(Pitch=0,Yaw=16384,Roll=0)
  End Object
  Components.Add(StaticMeshComponent4)

  //right clamp
  Begin Object class=StaticMeshComponent Name=StaticMeshComponent5
    StaticMesh=StaticMesh'NJS_LabAssets.MassSpringSystem.Clamp'
	LightEnvironment=MyLightEnvironment
	bUsePrecomputedShadows=FALSE
    scale = 0.5
	Translation=(X=216.0,Y=0.0,Z=52)
	Rotation=(Pitch=0,Yaw=-16384,Roll=0)
  End Object
  Components.Add(StaticMeshComponent5)

  //left stand
  Begin Object class=StaticMeshComponent Name=StaticMeshComponent6
    StaticMesh=StaticMesh'NJS_LabAssets.MassSpringSystem.Stand'
	LightEnvironment=MyLightEnvironment
	bUsePrecomputedShadows=FALSE
	Scale3D=(X=0.5,Y=0.5,Z=0.325)
	Translation=(X=-310.0,Y=0.0,Z=-40)
	Rotation=(Pitch=0,Yaw=16384,Roll=0)
  End Object
  Components.Add(StaticMeshComponent6)

  //Right stand
  Begin Object class=StaticMeshComponent Name=StaticMeshComponent7
    StaticMesh=StaticMesh'NJS_LabAssets.MassSpringSystem.Stand'
	LightEnvironment=MyLightEnvironment
	bUsePrecomputedShadows=FALSE
	Scale3D=(X=0.5,Y=0.5,Z=0.325)
	Translation=(X=310.0,Y=0.0,Z=-40)
	Rotation=(Pitch=0,Yaw=-16384,Roll=0)
  End Object
  Components.Add(StaticMeshComponent7)

  airtrack=StaticMeshComponent0      // airTrack
  glider = StaticMeshComponent1
  springLeft = StaticMeshComponent2
  springRight = StaticMeshComponent3
  clampLeft = StaticMeshComponent4
  clampRight = StaticMeshComponent5

  //accept traces
  bCollideActors=true
  bCollideWorld = true
  bBlockActors=true

  //custom var defaults
  bRunning=true
  mmass=1.0
  scaling=52.5
  deltaT=0.01
  stiffnessX=9.86960
  initX=0
  initVx=0
  desX=1.0
  damping=6

  Kp=50
  Ki=0
  Kd=0

}